clc;
clear all;
close all;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Parameters %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Memory Constraint. It is expressed in bytes. If no constraint on memory
% is desired, that value should be Inf.
memory = 1e7;
bytes_per_parameter = 4;
% Computational Constraint. It is expressed in number of operations. If no
% constraint on computation is desired, that value should be Inf.
computation = Inf;
% CNNs folder and filenames. The array of names ends with the string '\0'.
cnn_folder = 'to_be_fulfilled';
cnn_names = ["imagenet-caffe-alex.mat","imagenet-vgg-verydeep-16.mat","imagenet-vgg-verydeep-19.mat","\0"];
% The adopted classification algorithm is a list of strings ending with
% '\0'. The available algorithms are:
%       - Support Vector Machines ('SVM') with linear or Gaussian kernels.
classification_algorithms = ["linearSVM","gaussianSVM", "\0"];
% Dataset Folder. The dataset is supposed to be organized in one folder per
% each class
datasetFolder = 'to_be_fulfilled';
% Name of folders (one per class) within the dataset folder to be
% considered.
categories = {'folder_class_1', 'folder_class_2', 'folder_class_3'};

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%% Methodology %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% The methodology loads and processes one CNN at a time.
for i=1:size(cnn_names,2)-1
    fprintf('Working with the following CNN %s\n',char(cnn_names(i)));
    
    %%%%%%% Currently it works only with Beta-17 non-DAG MatConvNet CNNs %%%%%%%%
    convnet = cell(3,1);
    % cnn 32-bit
    convnet{1} = helperImportMatConvNet(fullfile(cnn_folder,char(cnn_names(i))));
    % cnn 16-bit
    convnet{2} = helperImportMatConvNetChangeWeights(fullfile(cnn_folder,char(cnn_names(i))),4,7,'truncation');
    % cnn 8-bit
    convnet{3} = helperImportMatConvNetChangeWeights(fullfile(cnn_folder,char(cnn_names(i))),5,7,'truncation');
    
    % At this point the methodology incrementally explores the given CNN,
    % until it overcomes either the memory or the time constraint.
    
    % Call the cnnInspection function, that has the role of incrementally
    % explores the CNN to compute the layers' structure and both the memory
    % and computation requirements.
    [l_name, l_memory, l_computation, imageSize] = ...
                            cnnInspection(convnet{1}, bytes_per_parameter);

    % Analysis with 8, 16 and 32-bit data types.
    % The 16-bit data type halves the required memory, the 8-bit requires a
    % fourth of memory. This is taken into account when computing the
    % memory constaints.
    for j=1:3
        fprintf('Analysis with a %d-bit datatype (No precision scaling).\n', 2^(6-j));
        
        % Compute the levels that satisfy both the memory and the computational
        % constraints.
        l_ok = (l_memory < memory*(2^(j-1))) & (l_computation < computation);
        [min_value,idx_last_level] = min(l_ok);
        if min_value==1
           fprintf('The entire CNN satisfies the constraints.\n'); 
        else
            if (idx_last_level<=2)
                fprintf('No approximation is feasible with these constraints\n');
            else
                fprintf('With this CNN the following levels are reachable without violating any constraints\n');
                disp(l_name(1:idx_last_level-1));

                % Compute the last pooling or fully-connected layer
                featureLayer = computeLastLayer(l_name(1:idx_last_level-1));
                %disp(featureLayer);

                for aa=1:size(classification_algorithms,2)-1
                
                    [accuracy, ~] = estimateAccuracy(convnet{i}, ...
                                    imageSize, datasetFolder, categories, ...
                                    char(featureLayer), classification_algorithms(aa));
                	fprintf('Considering the feasible layer %s, the estimated accuracy with %s is %.04f\n\n', ...
                            featureLayer, classification_algorithms(aa), accuracy);
                end
            end
            %disp(estimateAccuracy(cnn32,datasetFolder,'SVM'));
        end
    end
end

function label = computeLastLayer(l)
    l = flip(l);
    
    validLayers = double(contains(l,["pool","fc"],'IgnoreCase',true));
    [~, idx] = max(validLayers);
    
    label = l(idx);
end
