function [accuracy, confusionMatrix] = estimateAccuracy(convnet, ...
                                    netInputSize, datasetFolder, categories, ...
                                    featureLayer, algorithm)
    
    % Create a datastore from the collection of imeage data specified 
    % by the first parameter.
    imdsOriginal = imageDatastore(fullfile(datasetFolder, categories), ...
        'LabelSource', 'foldernames');
    % Count the number of images per each category, to then balance the
    % dataset.
    tbl = countEachLabel(imdsOriginal);
    minSetCount = min(tbl{:,2}); 
    
    imds = splitEachLabel(imdsOriginal, minSetCount, 'randomize');
    imds.ReadFcn = @(filename)readAndPreprocessImage(filename, netInputSize);
        
    % Split into training and test
    [trainingSet, testSet] = splitEachLabel(imds, 0.70, 'randomize');
    
    % Extract the training features
    trainingFeatures = activations(convnet, trainingSet, featureLayer, ...
            'MiniBatchSize', 32, 'OutputAs', 'columns');
    trainingLabels = trainingSet.Labels;
    
    % Fit the desired algorithm
    switch algorithm
        case 'linearSVM'
            t = templateSVM('Standardize',1,'KernelFunction','linear');
            classifier = fitcecoc(trainingFeatures, trainingLabels, ...
                'Learners', t, 'Coding', 'onevsall', ...
                'ObservationsIn', 'columns');
        case 'gaussianSVM'
            t = templateSVM('Standardize',1,'KernelFunction','gaussian');
            classifier = fitcecoc(trainingFeatures, trainingLabels, ...
                'Learners', t, 'Coding', 'onevsall', ...
                'ObservationsIn', 'columns');
        otherwise
            error('Unsupported algorithm!')
    end
    
    % Extract the test features and predict
    testFeatures = activations(convnet, testSet, featureLayer, ...
            'MiniBatchSize', 32);
    testLabels = testSet.Labels;
    
    predictedLabels = predict(classifier, testFeatures);
    
    % Compute confusion matrix and accuracy
    confusionMatrix = confusionmat(testLabels, predictedLabels);
    confusionMatrix = bsxfun(@rdivide, confusionMatrix, sum(confusionMatrix,2));
    accuracy = mean(diag(confusionMatrix));
    
end